use std::sync::Arc;

use anyhow::Result;
use dbus::{nonblock::SyncConnection, Path};
use serde_derive::{Deserialize, Serialize};

use crate::dbus_wrappers::{
    active_connection::ActiveConnectionWrapper, connection::ConnectionWrapper,
    device::DeviceWrapper, manager::ManagerWrapper, manager_settings::ManagerSettingsWrapper,
};

#[derive(Serialize, Deserialize, Clone, Debug)]
#[serde(tag = "type")]
pub enum DeviceIdentifier {
    DeviceInterface { device_interface: String },
    DevicePath { device_path: String },
}

impl DeviceIdentifier {
    pub async fn to_path(&self, conn: &Arc<SyncConnection>) -> Result<Path<'static>> {
        match self {
            DeviceIdentifier::DeviceInterface { device_interface } => {
                let manager = ManagerWrapper::from_connection(conn).await;
                let device_path = manager
                    .get_device_path_by_ip_iface(device_interface)
                    .await?;

                let owned_path = device_path.into_static();

                Ok(owned_path)
            }
            DeviceIdentifier::DevicePath { device_path } => Ok(Path::from(device_path.clone())),
        }
    }

    pub async fn to_device(
        &self,
        conn: &Arc<SyncConnection>,
    ) -> Result<Option<DeviceWrapper<'static>>> {
        match self.to_path(conn).await {
            Ok(device_path) => {
                let device = DeviceWrapper::from_path(conn.clone(), device_path).await;

                Ok(Some(device))
            }
            Err(err) => {
                log::warn!("Unable to resolve device identifier: {:?} {:?}", self, err);

                Ok(None)
            }
        }
    }
}

#[derive(Serialize, Deserialize, Clone, Debug)]
#[serde(tag = "type")]
pub enum ConnectionIdentifier {
    ConnectionPath { connection_path: String },
    ConnectionUUID { connection_uuid: String },
}

impl ConnectionIdentifier {
    pub async fn to_connection(
        &self,
        conn: &Arc<SyncConnection>,
    ) -> Result<Option<ConnectionWrapper<'_>>> {
        match self {
            ConnectionIdentifier::ConnectionPath { connection_path } => {
                let connection = ConnectionWrapper::from_path(connection_path).await;

                Ok(Some(connection))
            }
            ConnectionIdentifier::ConnectionUUID { connection_uuid } => {
                let manager_settings = ManagerSettingsWrapper::from_connection(conn).await;
                let connection = manager_settings
                    .get_connection_from_uuid(connection_uuid)
                    .await?;

                Ok(connection)
            }
        }
    }
}

#[derive(Serialize, Deserialize, Clone, Debug)]
#[serde(tag = "type")]
pub enum ActiveConnectionIdentifier {
    ConnectionUUID { connection_uuid: String },
    ConnectionPath { connection_path: String },
    DeviceInterface { device_interface: String },
}

impl ActiveConnectionIdentifier {
    pub async fn to_active_connection(
        &self,
        conn: &Arc<SyncConnection>,
    ) -> Result<Option<ActiveConnectionWrapper<'_>>> {
        match self {
            ActiveConnectionIdentifier::ConnectionUUID { connection_uuid } => {
                let manager = ManagerWrapper::from_connection(conn).await;
                let active_connection = manager
                    .get_active_connection_from_uuid(connection_uuid)
                    .await?;

                Ok(active_connection)
            }
            ActiveConnectionIdentifier::ConnectionPath { connection_path } => {
                let active_connection =
                    ActiveConnectionWrapper::from_path(conn.clone(), connection_path).await;

                Ok(Some(active_connection))
            }
            ActiveConnectionIdentifier::DeviceInterface { device_interface } => {
                let manager = ManagerWrapper::from_connection(conn).await;
                let device_path = manager
                    .get_device_path_by_ip_iface(device_interface)
                    .await?;

                let device = DeviceWrapper::from_path(conn.clone(), device_path).await;

                device.get_active_connection().await
            }
        }
    }
}
