#![deny(
    rust_2018_idioms,
    nonstandard_style,
    clippy::doc_markdown,
    clippy::match_bool,
    clippy::match_same_arms
)]
#![warn(future_incompatible)]

use collective::{
    cli::{AppOpts, ConfigurableAppOpts},
    config::ConfigFileFormat,
};
use dbus_tokio::connection;
use std::{path::PathBuf, sync::Arc};
use tokio::{
    signal,
    sync::{broadcast, mpsc},
};

use anyhow::Result;
use clap::Parser;

use crate::event::{handle_events, Event};

mod action;
mod condition;
mod config;
mod dbus_wrappers;
mod device_watcher;
mod event;
mod identifier;
mod rule;
mod watcher;

#[tokio::main]
async fn main() -> Result<()> {
    inner_main().await?;

    Ok(())
}

#[derive(Parser)]
#[clap(version = "1.0", author = "Eduardo T. <ed@trujillo.io>")]
struct Opts {
    /// Sets a custom config file.
    #[clap(short, long)]
    config: Option<PathBuf>,
    /// A level of verbosity, and can be used multiple times
    #[clap(short, long, parse(from_occurrences))]
    verbose: i32,
}

impl AppOpts for Opts {
    fn get_log_level_filter(&self) -> Option<log::LevelFilter> {
        match self.verbose {
            3 => Some(log::LevelFilter::Trace),
            2 => Some(log::LevelFilter::Debug),
            1 => Some(log::LevelFilter::Info),
            _ => None,
        }
    }
}

impl ConfigurableAppOpts<config::Config> for Opts {
    fn get_additional_config_paths(&self) -> Vec<(PathBuf, std::option::Option<ConfigFileFormat>)> {
        let mut paths = vec![];

        if let Some(config_from_opts) = &self.config {
          paths.push((config_from_opts.clone(), None));
        }

        let xdg_dirs = xdg::BaseDirectories::with_prefix("nm-reactor").unwrap();
        let xdg_config_path = xdg_dirs.find_config_file("config.yaml");

        if let Some(xdg_config_path) = xdg_config_path {
            paths.push((xdg_config_path, None));
        }

        if let Some(config_path) = &self.config {
            paths.push((config_path.clone(), None))
        }

        paths
    }

    fn get_config_file_format(&self) -> ConfigFileFormat {
        ConfigFileFormat::Yaml
    }
}

async fn inner_main() -> anyhow::Result<()> {
    let (_opts, config) = Opts::try_init_with_config()?;
    let config = Arc::new(config);

    log::info!("Parsed config and {} rules", config.rules.len());

    let (resource, conn) = connection::new_system_sync()?;

    log::info!("Connected to DBus");

    let _handle = tokio::spawn(async {
        let err = resource.await;
        panic!("Lost connection to D-Bus: {}", err);
    });

    let (stop_signal_tx, _stop_signal_rx) = broadcast::channel(1);
    let (event_tx, event_rx) = mpsc::channel(50);
    let lifecycle_event_tx = event_tx.clone();

    let mut task_handles = vec![];

    let watcher_handle = watcher::watch(&conn, stop_signal_tx.subscribe(), event_tx).await?;
    let event_handler_handle = handle_events(conn.clone(), config, event_rx).await?;

    task_handles.push(watcher_handle);
    task_handles.push(event_handler_handle);

    lifecycle_event_tx.send(Event::DaemonStarted).await?;

    signal::ctrl_c().await?;

    lifecycle_event_tx.send(Event::DaemonStopping).await?;

    stop_signal_tx.send(())?;

    for task_handle in task_handles {
        task_handle.await??;
    }

    Ok(())
}
