use std::sync::Arc;

use anyhow::Result;
use dbus::nonblock::SyncConnection;
use futures::stream::{self, StreamExt};
use serde_derive::{Deserialize, Serialize};

use crate::{
    action::Action,
    condition::Condition,
    event::{Event, Trigger},
};

#[derive(Serialize, Deserialize, Clone, Debug)]
pub struct Rule {
    // Defines which events should trigger this rule.
    //
    // Multiple triggers are evaluated using an OR operator.
    pub triggers: Vec<Trigger>,
    // Defines the conditions that should pass in order to perform any actions.
    //
    // Multiple conditions are evaluated using an AND operator.
    pub conditions: Vec<Condition>,
    // Defines actions that should be performed when the rule is triggered and
    // conditions pass.
    //
    // Actions are executed serially.
    pub actions: Vec<Action>,
}

impl Rule {
    pub async fn evaluate(&self, conn: &Arc<SyncConnection>, event: &Event) -> Result<()> {
        let any_trigger_matches = stream::iter(self.triggers.iter())
            .any(|trigger| async move {
                match trigger.matches_event(conn, event).await {
                    Ok(matches) => matches,
                    Err(err) => {
                        log::error!("Got an error while evaluating a trigger: {:?}", err);

                        false
                    }
                }
            })
            .await;

        if !any_trigger_matches {
            return Ok(());
        }

        let all_conditions_pass = stream::iter(self.conditions.iter())
            .all(|condition| async move {
                match condition.evaluate(conn).await {
                    Ok(passes) => passes,
                    Err(err) => {
                        log::error!("Got an error while evaluating a condition: {:?}", err);

                        false
                    }
                }
            })
            .await;

        if !all_conditions_pass {
            return Ok(());
        }

        for action in &self.actions {
            action.execute(conn).await?;
        }

        Ok(())
    }
}
