use std::net::Ipv4Addr;

use async_trait::async_trait;
use base64::Engine;
use http::{
    header::{AUTHORIZATION, HOST},
    Request, Response, Uri,
};
use hyper::Body;
use tokio::net::TcpSocket;

use crate::error::Error;

use super::LocalApiAdapter;

/// Client that connects to the local tailscaled over TCP with a password. This
/// is used on Windows and macOS when sandboxing is enabled.
#[derive(Clone)]
pub struct TcpWithPasswordAdapter {
    port: u16,
    password: String,
}

impl TcpWithPasswordAdapter {
    pub fn new<S: Into<String>>(port: u16, password: S) -> Self {
        TcpWithPasswordAdapter {
            port,
            password: password.into(),
        }
    }

    async fn request(&self, request: Request<Body>) -> Result<Response<Body>, Error> {
        let stream = TcpSocket::new_v4()?
            .connect((Ipv4Addr::LOCALHOST, self.port).into())
            .await?;
        let (mut request_sender, connection) = hyper::client::conn::handshake(stream).await?;

        tokio::spawn(async move {
            if let Err(e) = connection.await {
                eprintln!("Error in connection: {}", e);
            }
        });

        let response = request_sender.send_request(request).await?;
        if response.status() == 200 {
            Ok(response)
        } else {
            Err(Error::UnprocessableEntity)
        }
    }

    fn get_request_builder(&self) -> http::request::Builder {
        Request::builder()
            .header(HOST, "local-tailscaled.sock")
            .header(
                AUTHORIZATION,
                format!(
                    "Basic {}",
                    base64::engine::general_purpose::STANDARD_NO_PAD
                        .encode(format!(":{}", self.password))
                ),
            )
    }
}

#[async_trait]
impl LocalApiAdapter for TcpWithPasswordAdapter {
    async fn get(&self, uri: Uri) -> Result<Response<Body>, Error> {
        let request = self
            .get_request_builder()
            .method("GET")
            .uri(uri)
            .body(Body::empty())?;

        let response = self.request(request).await?;
        Ok(response)
    }

    async fn patch(&self, uri: Uri, body: Body) -> Result<Response<Body>, Error> {
        let request = self
            .get_request_builder()
            .method("PATCH")
            .uri(uri)
            .body(body)?;

        let response = self.request(request).await?;
        Ok(response)
    }
}
