use std::path::{Path, PathBuf};

use async_trait::async_trait;
use http::{header::HOST, Request, Response, Uri};
use hyper::Body;
use tokio::net::UnixStream;

use crate::error::Error;

use super::LocalApiAdapter;

/// Client that connects to the local tailscaled over a unix socket. This is
/// used on Linux and other Unix-like systems.
#[derive(Clone)]
pub struct UnixStreamAdapter {
    socket_path: PathBuf,
}

impl UnixStreamAdapter {
    pub fn new<P: AsRef<Path>>(socket_path: P) -> Self {
        let socket_path = socket_path.as_ref().to_path_buf();

        UnixStreamAdapter { socket_path }
    }

    async fn request(&self, request: Request<Body>) -> Result<Response<Body>, Error> {
        let stream = UnixStream::connect(&self.socket_path).await?;
        let (mut request_sender, connection) = hyper::client::conn::handshake(stream).await?;

        tokio::spawn(async move {
            if let Err(e) = connection.await {
                eprintln!("Error in connection: {}", e);
            }
        });

        let response = request_sender.send_request(request).await?;
        if response.status() == 200 {
            Ok(response)
        } else {
            Err(Error::UnprocessableEntity)
        }
    }

    fn get_request_builder(&self) -> http::request::Builder {
        Request::builder().header(HOST, "local-tailscaled.sock")
    }
}

#[async_trait]
impl LocalApiAdapter for UnixStreamAdapter {
    async fn get(&self, uri: Uri) -> Result<Response<Body>, Error> {
        let request = self
            .get_request_builder()
            .method("GET")
            .uri(uri)
            .body(Body::empty())?;

        let response = self.request(request).await?;
        Ok(response)
    }

    async fn patch(&self, uri: Uri, body: Body) -> Result<Response<Body>, Error> {
        let request = self
            .get_request_builder()
            .method("PATCH")
            .uri(uri)
            .body(body)?;

        let response = self.request(request).await?;
        Ok(response)
    }
}
