#![deny(rust_2018_idioms, nonstandard_style)]
#![warn(future_incompatible)]

use std::{io::Read, net::SocketAddr, path::Path};

use adapters::{tcp::TcpWithPasswordAdapter, unix::UnixStreamAdapter, LocalApiAdapter};
use error::Error;
use http::Uri;
use hyper::{body::Buf, Body};
pub use types::*;

pub mod adapters;
/// Definitions of types used in the tailscale API
pub mod error;
pub mod types;

/// Client for the local tailscaled socket
#[derive(Clone)]
pub struct LocalApi<T: LocalApiAdapter> {
    /// Path to the tailscaled socket
    client: T,
}

impl LocalApi<UnixStreamAdapter> {
    /// Create a new client for the local tailscaled from the path to the
    /// socket.
    pub fn new_with_socket_path<P: AsRef<Path>>(socket_path: P) -> Self {
        Self {
            client: UnixStreamAdapter::new(socket_path),
        }
    }
}

impl LocalApi<TcpWithPasswordAdapter> {
    /// Create a new client for the local tailscaled from the TCP port and
    /// password.
    pub fn new_with_port_and_password<S: Into<String>>(port: u16, password: S) -> Self {
        Self {
            client: TcpWithPasswordAdapter::new(port, password),
        }
    }
}

impl<T: LocalApiAdapter> LocalApi<T> {
    /// Get the certificate and key for a domain. The domain should be one of
    /// the valid domains for the local node.
    pub async fn certificate_pair(
        &self,
        domain: &str,
    ) -> Result<(PrivateKey, Vec<Certificate>), Error> {
        let response = self
            .client
            .get(
                format!("/localapi/v0/cert/{domain}?type=pair")
                    .parse()
                    .unwrap(),
            )
            .await?;

        let body = hyper::body::aggregate(response.into_body()).await?;
        let items = rustls_pemfile::read_all(&mut body.reader())?;
        let (certificates, mut private_keys) = items
            .into_iter()
            .map(|item| match item {
                rustls_pemfile::Item::ECKey(data)
                | rustls_pemfile::Item::PKCS8Key(data)
                | rustls_pemfile::Item::RSAKey(data) => Ok((false, data)),
                rustls_pemfile::Item::X509Certificate(data) => Ok((true, data)),
                _ => Err(Error::UnknownCertificateOrKey),
            })
            .collect::<Result<Vec<_>, Error>>()?
            .into_iter()
            .partition::<Vec<(bool, Vec<u8>)>, _>(|&(cert, _)| cert);

        let certificates = certificates
            .into_iter()
            .map(|(_, data)| Certificate(data))
            .collect();
        let (_, private_key_data) = private_keys.pop().ok_or(Error::UnknownCertificateOrKey)?;
        let private_key = PrivateKey(private_key_data);

        Ok((private_key, certificates))
    }

    /// Get the status of the local node.
    pub async fn status(&self) -> Result<Status, Error> {
        let response = self
            .client
            .get(Uri::from_static("/localapi/v0/status"))
            .await?;
        let body = hyper::body::aggregate(response.into_body()).await?;
        let status = serde_json::de::from_reader(body.reader())?;

        Ok(status)
    }

    pub async fn get_prefs(&self) -> Result<Prefs, Error> {
        let response = self
            .client
            .get(Uri::from_static("/localapi/v0/prefs"))
            .await?;
        let body = hyper::body::aggregate(response.into_body()).await?;

        let mut temp = String::new();
        body.reader().read_to_string(&mut temp)?;
        println!("{}", temp);

        let response = self
            .client
            .get(Uri::from_static("/localapi/v0/prefs"))
            .await?;
        let body = hyper::body::aggregate(response.into_body()).await?;

        let prefs = serde_json::de::from_reader(body.reader())?;

        Ok(prefs)
    }

    pub async fn edit_prefs(&self, prefs: &MaskedPrefs) -> Result<Prefs, Error> {
        let response = self
            .client
            .patch(
                Uri::from_static("/localapi/v0/prefs"),
                Body::from(serde_json::ser::to_string(prefs)?),
            )
            .await?;
        let body = hyper::body::aggregate(response.into_body()).await?;

        let prefs = serde_json::de::from_reader(body.reader())?;

        Ok(prefs)
    }

    /// Request whois information for an address in the tailnet.
    pub async fn whois(&self, address: SocketAddr) -> Result<Whois, Error> {
        let response = self
            .client
            .get(
                format!("/localapi/v0/whois?addr={address}")
                    .parse()
                    .unwrap(),
            )
            .await?;
        let body = hyper::body::aggregate(response.into_body()).await?;
        let whois = serde_json::de::from_reader(body.reader())?;

        Ok(whois)
    }
}
