use std::{
    collections::HashMap,
    net::{IpAddr, SocketAddr},
};

use chrono::{DateTime, Utc};
use serde::{Deserialize, Deserializer, Serialize};
use serde::de::{Error};
use serde_aux::prelude::*;

/// State of the backend
#[derive(Deserialize, Debug)]
#[non_exhaustive]
pub enum BackendState {
    NoState,
    NeedsLogin,
    NeedsMachineAuth,
    Stopped,
    Starting,
    Running,
}

/// Status of a peer
#[derive(Deserialize, Debug)]
#[serde(rename_all = "PascalCase")]
pub struct PeerStatus {
    #[serde(rename = "ID")]
    pub id: String,
    pub public_key: String,
    #[serde(rename = "HostName")]
    pub hostname: String,
    #[serde(rename = "DNSName")]
    pub dnsname: String,
    #[serde(rename = "OS")]
    pub os: String,
    #[serde(rename = "UserID")]
    pub user_id: i64,
    #[serde(
        rename = "TailscaleIPs",
        deserialize_with = "deserialize_default_from_null"
    )]
    pub tailscale_ips: Vec<IpAddr>,
    #[serde(default, deserialize_with = "deserialize_default_from_null")]
    pub tags: Vec<String>,
    #[serde(default, deserialize_with = "deserialize_default_from_null")]
    pub primary_routes: Vec<String>,
    #[serde(deserialize_with = "deserialize_default_from_null")]
    pub addrs: Vec<String>,
    pub cur_addr: String,
    pub relay: String,
    pub rx_bytes: i64,
    pub tx_bytes: i64,
    pub created: DateTime<Utc>,
    pub last_write: DateTime<Utc>,
    pub last_seen: DateTime<Utc>,
    pub last_handshake: DateTime<Utc>,
    pub online: bool,
    #[serde(default)]
    pub keep_alive: bool,
    pub exit_node: bool,
    pub exit_node_option: bool,
    pub active: bool,
    #[serde(
        rename = "PeerAPIURL",
        deserialize_with = "deserialize_default_from_null"
    )]
    pub peer_api_url: Vec<String>,
    #[serde(default, deserialize_with = "deserialize_default_from_null")]
    pub capabilities: Vec<String>,
    #[serde(
        default,
        rename = "sshHostKeys",
        deserialize_with = "deserialize_default_from_null"
    )]
    pub ssh_hostkeys: Vec<String>,
    #[serde(default)]
    pub sharee_node: bool,
    pub in_network_map: bool,
    pub in_magic_sock: bool,
    pub in_engine: bool,
}

/// Status of the current tailnet.
#[derive(Deserialize, Debug)]
#[serde(rename_all = "PascalCase")]
pub struct TailnetStatus {
    pub name: String,
    #[serde(rename = "MagicDNSSuffix")]
    pub magic_dns_suffix: String,
    #[serde(rename = "MagicDNSEnabled")]
    pub magic_dns_enabled: bool,
}

/// Status of the local tailscaled.
#[derive(Deserialize, Debug)]
#[serde(rename_all = "PascalCase")]
pub struct Status {
    pub version: String,
    pub backend_state: BackendState,
    #[serde(rename = "AuthURL")]
    pub auth_url: String,
    #[serde(rename = "TailscaleIPs")]
    pub tailscale_ips: Vec<IpAddr>,
    #[serde(rename = "Self")]
    pub self_status: PeerStatus,
    #[serde(deserialize_with = "deserialize_default_from_null")]
    pub health: Vec<String>,
    pub current_tailnet: Option<TailnetStatus>,
    #[serde(deserialize_with = "deserialize_default_from_null")]
    pub cert_domains: Vec<String>,
    #[serde(deserialize_with = "deserialize_default_from_null")]
    pub peer: HashMap<String, PeerStatus>,
    pub user: HashMap<i64, UserProfile>,
}

/// Service protocol
#[derive(Deserialize, Debug, Copy, Clone)]
#[non_exhaustive]
pub enum ServiceProto {
    #[serde(rename = "tcp")]
    Tcp,
    #[serde(rename = "udp")]
    Udp,
    #[serde(rename = "peerapi4")]
    PeerAPI4,
    #[serde(rename = "peerapi6")]
    PeerAPI6,
    #[serde(rename = "peerapi-dns-proxy")]
    PeerAPIDNS,
}

/// Service running on a node
#[derive(Deserialize, Debug, Clone)]
#[serde(rename_all = "PascalCase")]
pub struct Service {
    pub proto: ServiceProto,
    pub port: u16,
    pub description: Option<String>,
}

/// Host information
#[derive(Deserialize, Debug, Clone)]
#[serde(rename_all = "PascalCase")]
pub struct Hostinfo {
    #[serde(rename = "OS")]
    pub os: Option<String>,
    #[serde(rename = "OSVersion")]
    pub os_version: Option<String>,
    pub hostname: Option<String>,
    pub services: Option<Vec<Service>>,
    #[serde(default, rename = "sshHostKeys")]
    pub ssh_hostkeys: Option<Vec<String>>,
}

/// Node in the tailnet
#[derive(Deserialize, Debug, Clone)]
#[serde(rename_all = "PascalCase")]
pub struct Node {
    #[serde(rename = "ID")]
    pub id: i64,
    #[serde(rename = "StableID")]
    pub stable_id: String,
    pub name: String,
    pub user: i64,
    pub sharer: Option<i64>,
    pub key: String,
    pub key_expiry: DateTime<Utc>,
    pub machine: String,
    pub disco_key: String,
    pub addresses: Vec<String>,
    #[serde(rename = "AllowedIPs")]
    pub allowed_ips: Vec<String>,
    pub endpoints: Option<Vec<SocketAddr>>,
    #[serde(rename = "DERP")]
    pub derp: Option<String>,
    pub hostinfo: Hostinfo,
    pub created: DateTime<Utc>,
    #[serde(default)]
    pub tags: Vec<String>,
    #[serde(default)]
    pub primary_routes: Vec<String>,
    pub last_seen: Option<DateTime<Utc>>,
    pub online: Option<bool>,
    pub keep_alive: Option<bool>,
    pub machine_authorized: Option<bool>, // TODO: Check the upstream code if this has changed to MachineStatus
    #[serde(default)]
    pub capabilities: Vec<String>,
    #[serde(deserialize_with = "deserialize_default_from_null")]
    pub computed_name: String,
    #[serde(deserialize_with = "deserialize_default_from_null")]
    pub computed_name_with_host: String,
}

/// User profile.
#[derive(Deserialize, Debug, Clone)]
#[serde(rename_all = "PascalCase")]
pub struct UserProfile {
    #[serde(rename = "ID")]
    pub id: i64,
    pub login_name: String,
    pub display_name: String,
    #[serde(rename = "ProfilePicURL")]
    pub profile_pic_url: String,
}

/// Whois response
#[derive(Deserialize, Debug, Clone)]
#[serde(rename_all = "PascalCase")]
pub struct Whois {
    pub node: Node,
    pub user_profile: UserProfile,
    #[serde(default)]
    pub caps: Vec<String>,
}

/// DER encoded X.509 certificate for the node. This can either be the leaf
/// certificate or part of the certificate chain.
pub struct Certificate(pub Vec<u8>);

/// DER encoded private key for the node
pub struct PrivateKey(pub Vec<u8>);

type NetfilterMode = u8;
type StableNodeID = String;

#[derive(Debug, Deserialize, Serialize, Clone, Default)]
#[serde(rename_all = "PascalCase")]
pub struct Prefs {
    // ControlURL is the URL of the control server to use.
    #[serde(rename = "ControlURL")]
    pub control_url: String,
    // RouteAll specifies whether to accept subnets advertised by other nodes on the Tailscale network.
    pub route_all: bool,
    // AllowSingleHosts specifies whether to install routes for each node IP on the tailscale network.
    pub allow_single_hosts: bool,
    // ExitNodeID and ExitNodeIP specify the node that should be used as an exit node for internet traffic.
    #[serde(rename = "ExitNodeID")]
    pub exit_node_id: Option<StableNodeID>,
    #[serde(rename = "ExitNodeIP", deserialize_with = "deserialize_ip_option")]
    pub exit_node_ip: Option<std::net::IpAddr>,
    // ExitNodeAllowLANAccess indicates whether locally accessible subnets should be routed directly or via the exit node.
    #[serde(rename = "ExitNodeAllowLANAccess")]
    pub exit_node_allow_lan_access: bool,
    // CorpDNS specifies whether to install the Tailscale network's DNS configuration, if it exists.
    #[serde(rename = "CorpDNS")]
    pub corp_dns: bool,
    // RunSSH is whether this node should run an SSH server.
    #[serde(rename = "RunSSH")]
    pub run_ssh: bool,
    // RunWebClient is whether this node should run a web client.
    pub run_web_client: Option<bool>,
    // WantRunning indicates whether networking should be active on this node.
    pub want_running: bool,
    // LoggedOut indicates whether the user intends to be logged out.
    pub logged_out: bool,
    // ShieldsUp indicates whether to block all incoming connections.
    pub shields_up: bool,
    // AdvertiseTags specifies groups that this node wants to join, for purposes of ACL enforcement.
    pub advertise_tags: Option<Vec<String>>,
    // Hostname is the hostname to use for identifying the node.
    pub hostname: String,
    // NotepadURLs is a debugging setting that opens OAuth URLs in notepad.exe on Windows.
    #[serde(rename = "NotepadURLs")]
    pub notepad_urls: bool,
    // ForceDaemon specifies whether a platform should keep running after the GUI ends and/or the user logs out.
    pub force_daemon: Option<bool>,
    // Egg is an optional debug flag.
    pub egg: Option<bool>,
    // The following block of options only have an effect on Linux.
    // AdvertiseRoutes specifies CIDR prefixes to advertise into the Tailscale network.
    pub advertise_routes: Option<Vec<cidr::IpCidr>>,
    // NoSNAT specifies whether to source NAT traffic going to destinations in AdvertiseRoutes.
    #[serde(rename = "NoSNAT")]
    pub no_snat: Option<bool>,
    // NetfilterMode specifies how much to manage netfilter rules for Tailscale.
    pub netfilter_mode: NetfilterMode,
    // OperatorUser is the local machine user name who is allowed to operate tailscaled without being root or using sudo.
    pub operator_user: Option<String>,
    // ProfileName is the desired name of the profile.
    pub profile_name: Option<String>,
    // AutoUpdate sets the auto-update preferences for the node agent.
    pub auto_update: Option<AutoUpdatePrefs>,
    // AppConnector sets the app connector preferences for the node agent.
    pub app_connector: Option<AppConnectorPrefs>,
    // PostureChecking enables the collection of information used for device posture checks.
    pub posture_checking: Option<bool>,
}

#[derive(Debug, Deserialize, Serialize, Clone, Default)]
#[serde(rename_all = "PascalCase")]
pub struct AutoUpdatePrefs {
    // Check specifies whether background checks for updates are enabled.
    check: bool,
    // Apply specifies whether background auto-updates are enabled.
    apply: bool,
}

#[derive(Debug, Deserialize, Serialize, Clone, Default)]
#[serde(rename_all = "PascalCase")]
pub struct AppConnectorPrefs {
    // Advertise specifies whether the app connector subsystem is advertising this node as a connector.
    advertise: bool,
}

#[derive(Debug, Deserialize, Serialize, Clone, Default)]
#[serde(rename_all = "PascalCase")]
pub struct MaskedPrefs {
    #[serde(flatten)]
    pub prefs: Prefs,

    #[serde(rename = "ControlURLSet")]
    pub control_url_set: Option<bool>,
    pub route_all_set: Option<bool>,
    pub allow_single_hosts_set: Option<bool>,
    #[serde(rename = "ExitNodeIDSet")]
    pub exit_node_id_set: Option<bool>,
    #[serde(rename = "ExitNodeIPSet")]
    pub exit_node_ip_set: Option<bool>,
    #[serde(rename = "ExitNodeAllowLANAccessSet")]
    pub exit_node_allow_lan_access_set: Option<bool>,
    #[serde(rename = "CorpDNSSet")]
    pub corp_dns_set: Option<bool>,
    #[serde(rename = "RunSSHSet")]
    pub run_ssh_set: Option<bool>,
    pub run_web_client_set: Option<bool>,
    pub want_running_set: Option<bool>,
    pub logged_out_set: Option<bool>,
    pub shields_up_set: Option<bool>,
    pub advertise_tags_set: Option<bool>,
    pub hostname_set: Option<bool>,
    #[serde(rename = "NotepadURLsSet")]
    pub notepad_urls_set: Option<bool>,
    pub force_daemon_set: Option<bool>,
    pub egg_set: Option<bool>,
    pub advertise_routes_set: Option<bool>,
    #[serde(rename = "NoSNATSet")]
    pub no_snat_set: Option<bool>,
    pub netfilter_mode_set: Option<bool>,
    pub operator_user_set: Option<bool>,
    pub profile_name_set: Option<bool>,
    pub auto_update_set: Option<bool>,
    pub app_connector_set: Option<bool>,
    pub posture_checking_set: Option<bool>,
}

fn deserialize_ip_option<'de, D>(deserializer: D) -> Result<Option<IpAddr>, D::Error>
where
    D: Deserializer<'de>,
{
    let s: String = Deserialize::deserialize(deserializer)?;
    if s.is_empty() {
        Ok(None)
    } else {
        match s.parse() {
            Ok(ip) => Ok(Some(ip)),
            Err(_) => Err(Error::custom("Error deserializng IP")),
        }
    }
}
